/*
 * agent_cmdu.c - cmdu building functions
 *
 * Copyright (C) 2020 IOPSYS Software Solutions AB. All rights reserved.
 *
 * Author: jakob.olsson@iopsys.eu
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>
#include <arpa/inet.h>
#include <sys/ioctl.h>
#include <net/if_arp.h>
#include <pthread.h>

#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif
#include <json-c/json.h>
#include <libubox/blobmsg.h>
#include <libubox/blobmsg_json.h>
#include <libubox/uloop.h>
#include <libubox/ustream.h>
#include <libubox/utils.h>
#include <libubus.h>

#include <timer_impl.h>
#include <cmdu.h>
#include <1905_tlvs.h>
#include <map2.h>
#include <i1905_wsc.h>
#include <map_module.h>

#include <uci.h>

#include "utils/utils.h"
#include "utils/debug.h"
#include "utils/liblist.h"
#include "steer_rules.h"
#include "config.h"
#include "agent.h"

#include "agent_tlv.h"
#include "agent_cmdu.h"
#include "cmdu_validate.h"


static void buildcmdu_cb(struct ubus_request *req,
			  int type, struct blob_attr *msg)
{
	struct cmdu_buff **buff = (struct cmdu_buff **) req->priv;
	struct blob_attr *tb[2];
	static const struct blobmsg_policy ev_attr[2] = {
		[0] = { .name = "type", .type = BLOBMSG_TYPE_INT32 },
		[1] = { .name = "data", .type = BLOBMSG_TYPE_STRING }
	};
	uint16_t cmdu_type = 0, mid = 0;
	char *data;
	uint8_t origin[6] = { 0 };
	uint8_t *tlv;
	uint32_t b_len;

	if (!msg)
		return;

	blobmsg_parse(ev_attr, 2, tb, blob_data(msg), blob_len(msg));

	if (!tb[0] || !tb[1])
		return;

	cmdu_type = blobmsg_get_u32(tb[0]);
	data = blobmsg_get_string(tb[1]);
	if (!data)
		return;

	dbg("|%s:%d| type = %u data = %s \n", __func__, __LINE__,
			cmdu_type, data);

	b_len = (strlen(data)/2) - 3;

	tlv = (uint8_t *) calloc(1, b_len);
	if (!tlv)
		return;

	strtob(data, b_len, tlv);

	*buff = cmdu_alloc_custom(cmdu_type, &mid, NULL, origin,
			tlv, b_len);
	free(tlv);
}

struct cmdu_buff *buildcmdu(struct agent *a, uint16_t msg_type)
{
	struct blob_buf b = { 0 };
	int ret = 0;
	uint32_t id;
	struct cmdu_buff *buff = NULL;

	dbg("|%s:%d| Entry\n", __func__, __LINE__);

	memset(&b, 0, sizeof(struct blob_buf));
	blob_buf_init(&b, 0);

	blobmsg_add_u32(&b, "type", (uint32_t)msg_type);

	if (ubus_lookup_id(a->ubus_ctx, "ieee1905", &id)) {
		dbg("|%s:%d| not present ieee1905", __func__, __LINE__);
		goto out;
	}

	ret = ubus_invoke(a->ubus_ctx, id, "buildcmdu",
			b.head, buildcmdu_cb,
			&buff, 5000);
	if (ret) {
		dbg("|%s:%d| ubus call failed for |ieee1905 buildcmdu|",
					__func__, __LINE__);
		goto out;
	}
out:
	blob_buf_free(&b);
	return buff;
}

struct cmdu_buff *agent_gen_ap_autoconfig_search(struct agent *a,
		struct wifi_radio_element *radio, uint8_t profile)
{
	struct cmdu_buff *frm = NULL;
	int ret = 0;
	uint8_t band = 0;
	uint16_t mid = 0;
	uint8_t origin[6] = {0x01, 0x80, 0xc2, 0x00, 0x00, 0x13};


	frm = cmdu_alloc_simple(CMDU_TYPE_AP_AUTOCONFIGURATION_SEARCH, &mid);
	if (!frm) {
		dbg("%s: -ENOMEM\n", __func__);
		return NULL;
	}

	CMDU_SET_RELAY_MCAST(frm->cdata);

	ret = agent_gen_al_mac(a, frm, a->almac);
	if (ret)
		goto out;

	ret = agent_gen_searched_role(a, frm, IEEE80211_ROLE_REGISTRAR);
	if (ret)
		goto out;

	band = wifi_band_to_ieee1905band(radio->band);
	trace("radio band = %d, band = %d\n", radio->band, band);

	ret = agent_gen_autoconf_freq_band(a, frm, band);
	if (ret)
		goto out;

	ret = agent_gen_supported_service(a, frm, SUPPORTED_SERVICE_MULTIAP_AGENT);
	if (ret)
		goto out;

	ret = agent_gen_searched_service(a, frm, SEARCHED_SERVICE_MULTIAP_CONTROLLER);
	if (ret)
		goto out;

	ret = agent_gen_map_profile(a, frm, MULTIAP_PROFILE_2);
	if (ret)
		goto out;

	memcpy(frm->origin, origin, 6);
	cmdu_put_eom(frm);

	dbg("%s:%d: radio = %s\n", __func__, __LINE__, radio->name);
	return frm;
out:
	cmdu_free(frm);
	return NULL;
}

struct cmdu_buff *agent_gen_ap_autoconfig_wsc(struct agent *a, struct cmdu_buff *rx_cmdu,
		struct wifi_radio_element *radio)
{
	trace("agent: %s: --->\n", __func__);
	struct cmdu_buff *frm = NULL;
	uint16_t mid = 0;
	int ret = 0;


	frm = cmdu_alloc_simple(CMDU_TYPE_AP_AUTOCONFIGURATION_WSC, &mid);
	if (!frm) {
		dbg("%s: -ENOMEM\n", __func__);
		return NULL;
	}

	memcpy(frm->origin, rx_cmdu->origin, 6);

	ret = agent_gen_wsc(a, frm, radio);
	if (ret)
		goto out;

	ret = agent_gen_ap_radio_adv_cap(a, frm, radio);
	if (ret)
		goto out;

	ret = agent_gen_ap_radio_basic_cap(a, frm, radio);
	if (ret)
		goto out;

	ret = agent_gen_profile2_ap_cap(a, frm);
	if (ret)
		goto out;

	trace("build ap autoconfig wsc!\n");
	cmdu_put_eom(frm);

	return frm;
out:
	cmdu_free(frm);
	return NULL;
}

struct cmdu_buff *agent_gen_ap_metrics_response(struct agent *a,
		struct cmdu_buff *rx_cmdu)
{
	struct cmdu_buff *frm;
	int ret;
	int i = 0, j = 0;
	uint16_t mid = 0;
	struct tlv *tv[2][16] = {0};
	struct tlv_policy a_policy[] = {
		[0] = {
			.type = MAP_TLV_AP_METRIC_QUERY,
			.present = TLV_PRESENT_ONE
		},
		[1] = {
			.type = MAP_TLV_AP_RADIO_IDENTIFIER,
			.present = TLV_PRESENT_OPTIONAL_MORE
		}
	};

	cmdu_parse_tlvs(rx_cmdu, tv, a_policy, 2);
	if (!tv[0][0])
		return NULL;

	mid = cmdu_get_mid(rx_cmdu);

	frm = cmdu_alloc_simple(CMDU_AP_METRICS_RESPONSE, &mid);
	if (!frm) {
		dbg("%s: -ENOMEM\n", __func__);
		return NULL;
	}

	memcpy(frm->origin, rx_cmdu->origin, 6);

	/* AP Radio Identifier TLV */
	while (tv[1][i]) {
		int radio_index;
		struct tlv_ap_radio_identifier *tmp =
			(struct tlv_ap_radio_identifier *) tv[1][i++]->data;

		radio_index = get_radio_index(a, tmp->radio);
		if (radio_index != -1) {
// #ifdef PROFILE2
			/* Radio Metrics TLV */
			ret = agent_gen_radio_metrics(a, frm, radio_index);
			if (ret)
				goto out;
// #endif
		}
	}

	/* AP Metric Query TLV */
	if (tv[0][0]) {
		struct wifi_radio_element *radio;
		struct wifi_bss_element *bss;
		struct tlv_ap_metric_query *tmp =
			(struct tlv_ap_metric_query *) tv[0][0]->data;

		for (j = 0; j < tmp->num_bss; j++) {
			int radio_index;
			int bss_index;
			struct sta *s;
			struct netif_fh *fh;
			struct agent_config_radio *rcfg;

			bss_index = get_radio_and_bss_index(a,
					tmp->bss[j].bssid,
					&radio_index);

			if (bss_index == -1)
				continue;

			radio = a->radios + radio_index;
			bss = radio->bsslist + bss_index;

			/* AP Metrics TLV */
			ret = agent_gen_ap_metrics(a, frm,
					radio_index, bss_index);
			if (ret)
				goto out;

// #ifdef PROFILE2
			/* AP Extended Metrics TLV */
			ret = agent_gen_ap_ext_metrics(a, frm,
					radio_index, bss_index);
			if (ret)
				goto out;
// #endif

			fh = wifi_get_netif_by_bssid(a, bss->bssid);
			if (fh == NULL)
				continue;

			rcfg = get_agent_config_radio(&a->cfg, fh->cfg->device);
			if (!rcfg)
				return NULL;

			list_for_each_entry(s, &fh->stalist, list) {
				if (rcfg->include_sta_stats) {
					/* Associated STA Traffic Stats TLV */
					ret = agent_gen_assoc_sta_traffic_stats(a, frm, s);
					if (ret)
						goto out;
				}

				if (rcfg->include_sta_metric) {
					/* Associated STA Link Metrics TLV */
					ret = agent_gen_assoc_sta_link_metrics(a, frm, s, bss->bssid);
					if (ret)
						goto out;
// #ifdef PROFILE2
					/* Associated STA Extended Link Metrics TLV */
					ret = agent_gen_assoc_sta_ext_link_metric(a, frm, s, bss->bssid);
					if (ret)
						goto out;
// #endif
				}

			}
		}
	}

	cmdu_put_eom(frm);

	return frm;
out:
	cmdu_free(frm);
	return NULL;
}

struct cmdu_buff *agent_gen_assoc_sta_metric_response_per_intf(
		struct agent *a, char *ifname)
{
	uint16_t mid = 0;
	struct cmdu_buff *frm;
	struct wifi_radio_element *radio;
	int i, ret;

	frm = cmdu_alloc_simple(CMDU_ASSOC_STA_LINK_METRICS_RESPONSE, &mid);
	if (!frm) {
		dbg("%s: -ENOMEM\n", __func__);
		return NULL;
	}

	memcpy(frm->origin, a->cntlr_almac, 6);

	radio = wifi_ifname_to_radio_element(a, ifname);
	if (!radio)
		goto error;

	for (i = 0; i < radio->num_bss; i++) {
		struct sta *s = NULL;
		struct netif_fh *fh;
		struct wifi_bss_element *bss;

		bss = radio->bsslist + i;
		fh = wifi_get_netif_by_bssid(a, bss->bssid);
		if (!fh)
			continue;

		list_for_each_entry(s, &fh->stalist, list) {

			ret = agent_gen_assoc_sta_link_metrics(a, frm,
					s, bss->bssid);
			if (ret)
				goto error;

// #ifdef PROFILE2
			ret = agent_gen_assoc_sta_ext_link_metric(a, frm,
					s, bss->bssid);
			if (ret)
				goto error;
// #endif
		}

	}

	cmdu_put_eom(frm);

	return frm;
error:
	cmdu_free(frm);

	return NULL;
}

struct cmdu_buff *agent_gen_assoc_sta_metric_responsex(struct agent *a,
		uint8_t *origin, struct sta *s, struct netif_fh *fh)
{
	uint16_t mid = 0;
	struct cmdu_buff *frm;
	int ret;

	frm = cmdu_alloc_simple(CMDU_ASSOC_STA_LINK_METRICS_RESPONSE, &mid);
	if (!frm) {
		dbg("%s: -ENOMEM\n", __func__);
		return NULL;
	}

	if (s) {
		/* Associated STA Link Metrics TLV */
		ret = agent_gen_assoc_sta_link_metrics(a, frm, s, fh->bssid);
		if (ret)
			goto error;

// #ifdef PROFILE2
		/* Associated STA Extended Link Metrics TLV */
		ret = agent_gen_assoc_sta_ext_link_metric(a, frm, s, fh->bssid);
		if (ret)
			goto error;
// #endif
	} else {
		uint8_t reason_code = 0x00;
		struct tlv *t;
		struct tlv_assoc_sta_link_metrics *data;

		/* Associated STA Link Metrics TLV */
		t = cmdu_reserve_tlv(frm, 20);
		if (!t)
			goto error;

		t->type = MAP_TLV_ASSOCIATED_STA_LINK_METRICS;
		t->len = sizeof(*data);

		data = (struct tlv_assoc_sta_link_metrics *) t->data;
		/* Reported BSS for STA */
		data->num_bss = 0;
		ret = cmdu_put_tlv(frm, t);
		if (ret) {
			dbg("%s: error: cmdu_put_tlv()\n", __func__);
			goto error;
		}

		/* Error Code TLV */
		reason_code = 0x02;	/* STA not associated with any BSS */
		ret = agent_gen_tlv_error_code(a, frm, data->macaddr,
				reason_code);
		if (ret)
			goto error;
	}

	cmdu_put_eom(frm);
	memcpy(frm->origin, origin, 6);
	return frm;
error:
	cmdu_free(frm);
	return NULL;
}

struct cmdu_buff *agent_gen_assoc_sta_metric_response(
		struct agent *a, struct cmdu_buff *rx_cmdu)
{
	uint16_t mid = 0;
	struct tlv *tv0;
	struct tlv_sta_mac *query;
	struct cmdu_buff *frm;
	struct tlv *tv[1][16];
	struct tlv_policy a_policy[] = {
		[0] = {
			.type = MAP_TLV_STA_MAC_ADDRESS,
			.present = TLV_PRESENT_ONE
		}
	};

	struct netif_fh *fh;
	struct sta *s;
	bool is_sta_found = false;
	int ret;

	ret = cmdu_parse_tlvs(rx_cmdu, tv, a_policy, 1);
	if (!tv[0][0])
		return NULL;

	mid = cmdu_get_mid(rx_cmdu);

	frm = cmdu_alloc_simple(CMDU_ASSOC_STA_LINK_METRICS_RESPONSE, &mid);
	if (!frm) {
		dbg("%s: -ENOMEM\n", __func__);
		return NULL;
	}

	memcpy(frm->origin, rx_cmdu->origin, 6);

	tv0 = tv[0][0];
	query = (struct tlv_sta_mac *) tv0->data;
	list_for_each_entry(fh, &a->fhlist, list) {
		list_for_each_entry(s, &fh->stalist, list) {
			if (hwaddr_equal(query->macaddr, s->macaddr)) {
				is_sta_found = true;
				break;
			}
		}
		if (is_sta_found)
			break;
	}

	if (is_sta_found) {
		/* Associated STA Link Metrics TLV */
		ret = agent_gen_assoc_sta_link_metrics(a, frm, s, fh->bssid);
		if (ret)
			goto error;

// #ifdef PROFILE2
		/* Associated STA Extended Link Metrics TLV */
		ret = agent_gen_assoc_sta_ext_link_metric(a, frm, s, fh->bssid);
		if (ret)
			goto error;
// #endif
	} else {
		uint8_t reason_code = 0x00;
		struct tlv *t;
		struct tlv_assoc_sta_link_metrics *data;

		/* Associated STA Link Metrics TLV */
		t = cmdu_reserve_tlv(frm, 20);
		if (!t)
			goto error;

		t->type = MAP_TLV_ASSOCIATED_STA_LINK_METRICS;
		t->len = sizeof(*data);

		data = (struct tlv_assoc_sta_link_metrics *) t->data;
		memcpy(data->macaddr, query->macaddr, 6);
		/* Reported BSS for STA */
		data->num_bss = 0;
		ret = cmdu_put_tlv(frm, t);
		if (ret) {
			dbg("%s: error: cmdu_put_tlv()\n", __func__);
			goto error;
		}

		/* Error Code TLV */
		reason_code = 0x02;	/* STA not associated with any BSS */
		ret = agent_gen_tlv_error_code(a, frm,
				query->macaddr, reason_code);
		if (ret)
			goto error;
	}

	cmdu_put_eom(frm);

	return frm;
error:
	cmdu_free(frm);

	return NULL;
}

struct cmdu_buff *agent_gen_beacon_metrics_query(struct agent *a,
			uint8_t *agent_mac, uint8_t *sta_addr, uint8_t opclass,
			uint8_t channel, uint8_t *bssid,
			uint8_t reporting_detail, char *ssid,
			uint8_t num_report, struct sta_channel_report *report,
			uint8_t num_element, uint8_t *element)
{
	struct cmdu_buff *frm = NULL;
	int ret = 0;
	uint16_t mid = 0;

	trace("agent: %s: --->\n", __func__);

	if (!agent_mac || !sta_addr || !bssid)
		return NULL;

	frm = cmdu_alloc_simple(CMDU_BEACON_METRICS_QUERY, &mid);
	if (!frm) {
		dbg("%s: -ENOMEM\n", __func__);
		return NULL;
	}

	/* Beacon metrics query TLV */
	ret = agent_gen_tlv_beacon_metrics_query(a, frm,
			sta_addr, opclass, channel, bssid,
			reporting_detail, ssid, num_report,
			report, num_element, element);

	if (ret)
		goto fail;

	/* destination agent */
	memcpy(frm->origin, agent_mac, 6);

	cmdu_put_eom(frm);

	return frm;
fail:
	cmdu_free(frm);
	return NULL;
}

struct cmdu_buff *agent_gen_cmdu_beacon_metrics_resp(struct agent *a,
			uint8_t *sta_addr, uint8_t report_elems_nr,
			uint8_t *report_elem, uint16_t elem_len)
{
	struct cmdu_buff *frm = NULL;
	uint8_t *tlv = NULL;
	uint16_t mid = 0;
	uint8_t origin[6] = { 0 };
	uint32_t t_len = 0;

	/* Since no excess allocation (1500 octets in cmdu_alloc_simple)
	 * and cmdu_put_tlv is not used - one must specify the exact size.
	 * I.e. must manually add +3: +1 for type and +2 for length.
	 */
	t_len = sizeof(struct tlv_beacon_metrics_resp) + elem_len + 3;

	tlv = (uint8_t *) calloc(1, t_len);

	if (!tlv)
		return NULL;

	if (agent_gen_tlv_beacon_metrics_resp(
					a, tlv, sta_addr,
					report_elems_nr,
					report_elem,
					elem_len))
		goto out;

	frm = cmdu_alloc_custom(CMDU_BEACON_METRICS_RESPONSE,
				&mid, NULL, origin, tlv, t_len);

	if (!frm) {
		dbg("%s: -ENOMEM\n", __func__);
		goto out;
	}

	/* destination controller */
	memcpy(frm->origin, a->cntlr_almac, 6);

	cmdu_put_eom(frm);
out:
	if (tlv)
		free(tlv);
	return frm;
}

struct cmdu_buff *agent_gen_unassoc_sta_metric_query(struct agent *a,
		uint8_t *origin, uint8_t opclass,
		uint8_t num_metrics, struct unassoc_sta_metric *metrics)
{
	int ret;
	uint16_t mid = 0;
	struct cmdu_buff *frm;

	/* TODO: A Multi-AP Agent shall not send an Unassociated
	 * STA Link Metrics Query message to a Multi-AP Agent that
	 * does not indicate support for Unassociated STA Link
	 * Metrics in the AP Capability TLV.
	 */

	frm = cmdu_alloc_simple(CMDU_UNASSOC_STA_LINK_METRIC_QUERY, &mid);
	if (!frm) {
		dbg("%s: -ENOMEM\n", __func__);
		return NULL;
	}

	memcpy(frm->origin, origin, 6);

	/* Unassociated STA link metrics query TLV */
	ret = agent_gen_tlv_unassoc_sta_lm_query(a, frm,
			opclass, num_metrics, metrics);

	if (ret)
		goto out;

	cmdu_put_eom(frm);
	return frm;

out:
	return NULL;
}

struct cmdu_buff *agent_gen_tunneled_msg(struct agent *a, uint8_t protocol,
		uint8_t *sta, int frame_len, uint8_t *frame_body)
{
	int ret;
	struct cmdu_buff *cmdu;
	uint16_t mid = 0;

	/* TODO: check profile type
	 * return NULL,in case of PROFILE-1
	 * #ifdef PROFILE1
	 * return NULL;
	 * #endif
	 */

	if (!sta && !frame_body)
		return NULL;

	cmdu = cmdu_alloc_simple(CMDU_TUNNELED, &mid);
	if (!cmdu) {
		err("%s: -ENOMEM\n", __func__);
		return NULL;
	}

	memcpy(cmdu->origin, a->cntlr_almac, 6);

	/* Source Info TLV */
	ret = agent_gen_source_info(a, cmdu, sta);
	if (ret)
		goto error;

	/* Tunneled message type TLV */
	ret = agent_gen_tunnel_msg_type(a, cmdu, protocol);
	if (ret)
		goto error;

	/* Tunneled TLV */
	ret = agent_gen_tunneled(a, cmdu, frame_len, frame_body);
	if (ret)
		goto error;

	cmdu_put_eom(cmdu);

	return cmdu;

error:
	cmdu_free(cmdu);

	return NULL;
}


struct cmdu_buff *agent_gen_vendor_specific_cmdu(struct agent *a,
		uint8_t *origin, uint8_t depth)
{
	return NULL;
//	struct tlv_vendor_specific *p;
//	struct cmdu_cstruct *cmdu;
//	int tlv_index = 0;
//
//	cmdu = calloc(1, sizeof(struct cmdu_cstruct));
//	if (!cmdu)
//		return NULL;
//
//	cmdu->message_type = CMDU_TYPE_VENDOR_SPECIFIC;
//	memcpy(cmdu->origin, origin, 6);
//	strncpy(cmdu->intf_name, a->cfg.al_bridge, IFNAMESIZE - 1);
//
//	p = agent_gen_vendor_specific_tlv(a, depth);
//	if (!p)
//		goto error;
//
//	cmdu->num_tlvs++;
//	cmdu->tlvs = (uint8_t **)calloc(cmdu->num_tlvs,
//			sizeof(uint8_t *));
//	cmdu->tlvs[tlv_index++] = (uint8_t *)p;
//	return cmdu;
//error:
//	map_free_cmdu(cmdu);
//	return NULL;
}

struct cmdu_buff *agent_gen_ch_scan_response(struct agent *a,
		struct cmdu_buff *cmdu)
{
	trace("%s --->\n", __func__);
	struct wifi_radio_element *radio = a->radios;
	int i, j, k, m, ret;
	int radio_index, opclass_index, channel_index;
	uint8_t query_channel; /* The current channel from the request TLV */
	int offset = 0;
	uint16_t mid = 0;
	struct cmdu_buff *cmdu_data; /* The response cmdu */
	struct tlv *tv[1][16];
	uint8_t *q_data = NULL;
	struct tlv_channel_scan_request *query = NULL;

	if (!validate_channel_scan_request(cmdu, tv)) {
		dbg("cmdu validation: [CHANNEL_SCAN_REQ] failed\n");
		return NULL;
	}

	q_data = (uint8_t *)tv[0][0]->data;
	query = (struct tlv_channel_scan_request *)q_data;


	/* Allocate the cmdu */
	cmdu_data = cmdu_alloc_frame(0x10000);
	if (!cmdu_data) {
		trace("%s: -ENOMEM\n", __func__);
		return NULL;
	}
	/* Define the cmdu */
	mid = cmdu_get_mid(cmdu);
	cmdu_set_type(cmdu_data, CMDU_CHANNEL_SCAN_REPORT);
	cmdu_set_mid(cmdu_data, mid);

	memcpy(cmdu_data->origin, cmdu->origin, 6);

	/* Define the TLVs */
	/* Timestamp TLV */
	ret = agent_gen_timestamp_tlv(a, cmdu_data);
	if (ret)
		goto error;
	trace("|%s:%d| Added TIMESTAMP_TLV\n", __func__, __LINE__);

	/* Channel scan result */
	offset = 1 + 1;

	for (m = 0; m < query->num_radio; m++) {
		struct channel_scan_request_radio *r =
				(struct channel_scan_request_radio *)&q_data[offset];
		// dbg("\tMID: %d Radio_id in %d iteration:" MACFMT "\n", mid, m, MAC2STR(r->radio));
		offset += sizeof(*r);
		radio_index = get_radio_index(a, r->radio);
		if (radio_index == -1)
				continue;
		struct channel_scan_request_opclass *op =
				(struct channel_scan_request_opclass *)&q_data[offset];
		offset += sizeof(*op);
		/* Success scan */
		if (a->status_code == CH_SCAN_STATUS_SUCCESS) {
			/* All channels scan */
			if (r->num_opclass == 0 || op->num_channel == 0) {
				for (i = 0; i < radio[radio_index].scanlist->num_opclass_scanned; i++) {
					if (radio[radio_index].scanlist->opclass_scanlist[i].bandwidth == 20) {
						for (j = 0; j < radio[radio_index].scanlist->opclass_scanlist[i].num_channels_scanned; j++) {
							opclass_index = i;
							channel_index = j;
							ret = agent_gen_ch_scan_respons_tlv(a, cmdu_data, radio_index, opclass_index, channel_index);
							if (ret)
								goto error;
							trace("|%s:%d| Added RESULT_TLV.\n", __func__, __LINE__);
						}
					}
				}
			/* Specified channels */
			} else {
				// trace("|%s:%d| The number of channels in the request %d \n", __func__, __LINE__, op->num_channel);
				for (k = 0; k < op->num_channel; k++) {
					query_channel = op->channel[k];
					for (i = 0; i < radio[radio_index].scanlist->num_opclass_scanned; i++) {
						for (j = 0; j < radio[radio_index].scanlist->opclass_scanlist[i].num_channels_scanned; j++) {
							if (radio[radio_index].scanlist->opclass_scanlist[i].channel_scanlist[j].channel == query_channel) {
								opclass_index = i;
								channel_index = j;
								ret = agent_gen_ch_scan_respons_tlv(a, cmdu_data, radio_index, opclass_index, channel_index);
								if (ret)
									goto error;
								trace("|%s:%d| Added RESULT_TLV.\n", __func__, __LINE__);
								goto found;
							}
						}
					}
found:
					continue;
				}
				offset += op->num_channel;
			}
		/* Non-success scan */
		} else {
			for (k = 0; k < op->num_channel; k++) {
				query_channel = op->channel[k];
				offset++;
				ret = agent_gen_non_success_ch_scan_respons_tlv(a, cmdu_data, radio_index, 0, query_channel);
				if (ret)
					goto error;
				trace("|%s:%d| Added RESULT_TLV.\n", __func__, __LINE__);
			}
		}
	}
	cmdu_put_eom(cmdu_data);
	return cmdu_data;
error:
	dbg("|%s:%d| out of memory!\n", __func__, __LINE__);
	cmdu_free(cmdu_data);
	return NULL;
}

struct cmdu_buff *agent_gen_ap_caps_query(struct agent *a, uint8_t *origin)
{
	uint16_t mid = 0;
	struct cmdu_buff *resp;

	/* Allocate the cmdu_data structure */
	resp = cmdu_alloc_simple(CMDU_AP_CAPABILITY_QUERY, &mid);
	if (!resp) {
		dbg("%s: -ENOMEM\n", __func__);
		return NULL;
	}

	memcpy(resp->origin, origin, 6);
	cmdu_put_eom(resp);
	return resp;
}

struct cmdu_buff *agent_gen_ap_caps_response(struct agent *a,
		struct cmdu_buff *rec_cmdu)
{
	int i, ret;
	uint16_t mid = 0;
	struct cmdu_buff *resp;
	struct wifi_radio_element *radio;

	mid = cmdu_get_mid(rec_cmdu);
	resp = cmdu_alloc_simple(CMDU_AP_CAPABILITY_REPORT, &mid);
	if (!resp) {
		fprintf(stderr, "%s: -ENOMEM\n", __func__);
		return NULL;
	}
	memcpy(resp->origin, rec_cmdu->origin, 6);

	/* AP Capability TLV */
	ret = agent_gen_ap_caps(a, resp);
	if (ret)
		goto error;

	/* AP Radio Basic Capabilities TLV */
	for (i = 0; i < a->num_radios; i++) {
		radio = a->radios + i;
		ret = agent_gen_ap_radio_basic_cap(a, resp, radio);
		if (ret)
			goto error;
	}

	/* AP HT Capabilities TLV */
	for (i = 0; i < a->num_radios; i++) {
		ret = agent_gen_ap_ht_caps(a, resp, i);
		if (ret)
			goto error;
	}

	/* AP VHT Capabilities TLV */
	for (i = 0; i < a->num_radios; i++) {
		ret = agent_gen_ap_vht_caps(a, resp, i);
		if (ret)
			goto error;
	}

	/* AP HE Capabilities TLV */
	for (i = 0; i < a->num_radios; i++) {
		radio = a->radios + i;
		ret = agent_gen_ap_he_caps(a, resp, i);
		if (ret)
			goto error;
	}

	if (a->cfg.profile == 2) {
		/* Channel Scan Capabilities TLV */
		ret = agent_gen_ch_scan_cap(a, resp);
		if (ret)
			goto error;

		/* CAC Capabilities TLV */
		ret = agent_gen_cac_cap(a, resp);
		if (ret)
			goto error;

		/* Profile-2 AP Capability TLV */
		ret = agent_gen_profile2_ap_cap(a, resp);
		if (ret)
			goto error;

		/* Metric Collection Interval TLV */
		ret = agent_gen_metric_collection_interval(a, resp);
		if (ret)
			goto error;
	}

	cmdu_put_eom(resp);

	return resp;

error:
	cmdu_free(resp);

	return NULL;
}

struct cmdu_buff *agent_gen_bk_caps_response(struct agent *a,
		struct cmdu_buff *cmdu)
{
	struct cmdu_buff *cmdu_data;
	uint32_t i;
	uint16_t mid;
	int ret;

	/* Allocate and define the cmdu */
	mid = cmdu_get_mid(cmdu);
	cmdu_data = cmdu_alloc_simple(CMDU_BACKHAUL_STA_CAPABILITY_REPORT, &mid);
	if (!cmdu_data) {
		fprintf(stderr, "%s: -ENOMEM\n", __func__);
		return NULL;
	}
	strncpy(cmdu_data->dev_ifname, cmdu->dev_ifname, 16);
	memcpy(cmdu_data->origin, cmdu->origin, 6);

	/* Define the TLVs */
	for (i = 0; i < a->num_radios; i++) {
		/* Backhaul STA radio capabilities TLV */
		ret = agent_gen_bk_sta_radio_cap_tlv(a, i, cmdu_data);
		if (ret)
			goto error;
		trace("|%s:%d| Added MAP_TLV_BACKHAUL_STA_RADIO_CAPABILITY\n", __func__, __LINE__);
	}
	cmdu_put_eom(cmdu_data);
	return cmdu_data;
error:
	cmdu_free(cmdu_data);
	return NULL;
}

struct cmdu_buff *agent_gen_topology_notification(struct agent *agent,
		uint8_t *mac, uint8_t *bssid, uint8_t assoc_event)
{

	struct cmdu_buff *frm = NULL;
	int ret = 0;
	uint16_t mid = 0;
	uint8_t origin[6] = {0x01, 0x80, 0xc2, 0x00, 0x00, 0x13};


	frm = cmdu_alloc_simple(CMDU_TYPE_TOPOLOGY_NOTIFICATION, &mid);
	if (!frm) {
		dbg("%s: -ENOMEM\n", __func__);
		return NULL;
	}

	CMDU_SET_RELAY_MCAST(frm->cdata);

	ret = agent_gen_al_mac(agent, frm, agent->almac);
	if (ret)
		goto out;

	/* Client Association Event TLV */
	ret = agent_gen_client_assoc_event_tlv(agent, frm, mac, bssid,
			assoc_event);
	if (ret)
		goto out;

	memcpy(frm->origin, origin, 6);
	cmdu_put_eom(frm);
	return frm;
out:
	cmdu_free(frm);
	return NULL;
}

struct cmdu_buff *agent_gen_client_disassoc(struct agent *a,
		uint8_t *mac, uint8_t *bssid)
{
	uint16_t mid = 0;
	int ret;
	struct cmdu_buff *cmdu;
	struct sta *s;

	/* TODO/hardcoded
	 * need this data from wifi.sta event
	 */
	uint16_t reason_code = 0x00;

	cmdu = cmdu_alloc_simple(CMDU_CLIENT_DISASSOCIATION_STATS, &mid);
	if (!cmdu) {
		dbg("%s: -ENOMEM\n", __func__);
		return NULL;
	}

	memcpy(cmdu->origin, a->cntlr_almac, 6);

	s = find_sta_by_mac(a, mac);
	if (!s) {
		dbg("sta " MACFMT " is not associated with any BSS\n",
				MAC2STR(mac));
		goto error;
	}

	/* STA MAC Address TLV */
	ret = agent_gen_sta_mac(a, cmdu, mac);
	if (ret)
		goto error;

	/* Reason Code TLV */
	ret = agent_gen_reason_code(a, cmdu, reason_code);
	if (ret)
		goto error;

	/* Associated STA Traffic Stats TLV */
	ret = agent_gen_assoc_sta_traffic_stats(a, cmdu, s);
	if (ret)
		goto error;

	cmdu_put_eom(cmdu);

	return cmdu;
error:
	cmdu_free(cmdu);

	return NULL;
}

struct cmdu_buff *agent_gen_topology_query(struct agent *a, uint8_t *origin)
{
	struct cmdu_buff *resp;
	uint16_t mid = 0;

	resp = cmdu_alloc_simple(CMDU_TYPE_TOPOLOGY_QUERY, &mid);
	if (!resp) {
		dbg("%s: -ENOMEM\n", __func__);
		return NULL;
	}

	memcpy(resp->origin, origin, 6);
	cmdu_put_eom(resp);
	return resp;
}

struct cmdu_buff *agent_gen_topology_response(struct agent *a, uint8_t *origin,
	uint16_t mid)
{
	struct cmdu_buff *resp = NULL, *ext = NULL;
	int ret;

	/* query i1905d base CMDU */
	resp = buildcmdu(a, CMDU_TYPE_TOPOLOGY_RESPONSE);
	if (!resp) {
		dbg("No response from stack when generating 0x%04x\n",
				CMDU_TYPE_TOPOLOGY_RESPONSE);
		return NULL;
	}

	ext = cmdu_realloc(resp, 2000);
	if (!ext)
		goto error;

	resp = ext;

	/* Supported Service */
	ret = agent_gen_supported_service(a, resp,
			(!memcmp(a->almac, a->cntlr_almac, 6) ?
			SUPPORTED_SERVICE_MULTIAP_CONTROLLER :
			SUPPORTED_SERVICE_MULTIAP_AGENT));
	if (ret)
		goto error;

	/* AP Operational BSS */
	ret = agent_gen_ap_oper_bss_tlv(a, resp);
	if (ret)
		goto error;

	/* Associated Clients */
	ret = agent_gen_assoc_client_tlv(a, resp);
	if (ret)
		goto error;

#ifndef EASYMESH_R2_CERT
	/* Vendor Specific tlv added to get the sta link metrics in host.devices*/
	ret = agent_gen_vendor_specific_sta_linkmetric_tlv(a, resp);
	if (ret)
		goto error;

	/* Vendor Specific tlv added to get the sta link metrics in host.devices*/
	ret = agent_gen_vendor_specific_bbbs_tlv(a, resp);
	if (ret)
		goto error;
#endif

	/* Multi-AP profile */
	ret = agent_gen_map_profile(a, resp, MULTIAP_PROFILE_2);
	if (ret)
		goto error;

	cmdu_set_mid(resp, mid);
	memcpy(resp->origin, origin, 6);
	cmdu_put_eom(resp);
	return resp;
error:
	cmdu_free(resp);
	return NULL;
}

struct cmdu_buff *agent_gen_cmdu_1905_ack(
		struct agent *a, uint8_t *origin, uint16_t mid,
		struct sta_error_response *sta_resp, uint32_t sta_count)
{
	trace("agent: %s: --->\n", __func__);
	struct cmdu_buff *frm = NULL;
	int j;

	frm = cmdu_alloc_simple(CMDU_1905_ACK, &mid);
	if (!frm) {
		dbg("%s: -ENOMEM\n", __func__);
		return NULL;
	}

	memcpy(frm->origin, origin, 6);

	/* Error Code TLV 17.2.36 */
	for (j = 0; j < sta_count; j++) {
		if (agent_gen_tlv_error_code(a, frm,
				sta_resp[j].sta_mac, sta_resp[j].response))
			continue;

	}

	cmdu_put_eom(frm);
	return frm;
}

struct cmdu_buff *agent_gen_cmdu_backhaul_steer_resp(struct agent *a,
		uint8_t *target_bssid, uint8_t *macaddr, uint16_t mid)
{
	trace("agent: %s: --->\n", __func__);
	struct cmdu_buff *frm = NULL;
	int ret = 0;


	frm = cmdu_alloc_simple(CMDU_BACKHAUL_STEER_RESPONSE, &mid);
	if (!frm) {
		dbg("%s: -ENOMEM\n", __func__);
		return NULL;
	}

	ret = agent_gen_tlv_backhaul_steer_resp(a, frm, target_bssid, macaddr);
	if (ret)
		goto out;

	cmdu_put_eom(frm);

	return frm;
out:
	cmdu_free(frm);
	return NULL;
}

struct cmdu_buff *agent_gen_channel_preference_report(struct agent *a,
		struct cmdu_buff *rx_cmdu)
{
	uint16_t mid = 0;
	int i, ret;
	struct cmdu_buff *frm;

	if (rx_cmdu)
		mid = cmdu_get_mid(rx_cmdu);

	frm = cmdu_alloc_simple(CMDU_CHANNEL_PREFERENCE_REPORT, &mid);
	if (!frm) {
		dbg("%s: -ENOMEM\n", __func__);
		return NULL;
	}

	if (rx_cmdu)
		memcpy(frm->origin, rx_cmdu->origin, 6);
	else
		memcpy(frm->origin, a->cntlr_almac, 6);

	for (i = 0; i < a->num_radios; i++) {
		/* Channel Preference TLV */
		ret = agent_gen_channel_pref(a, frm, i);
		if (ret)
			goto error;
	}

	for (i = 0; i < a->num_radios; i++) {
		/* Radio Operation Restriction TLV */
		ret = agent_gen_radio_oper_restrict(a, frm, i);
		if (ret)
			goto error;
	}

//#ifdef PROFILE2
	/* CAC Completion Report TLV */
	ret = agent_gen_cac_complete_report(a, frm);
	if (ret)
		goto error;

	/* CAC Status Report TLV */
	ret = agent_gen_cac_status_report(a, frm);
	if (ret)
		goto error;
//#endif

	cmdu_put_eom(frm);
	return frm;

error:
	cmdu_free(frm);
	return NULL;
}

struct cmdu_buff *agent_gen_association_status_notify(struct agent *a,
		int num_data, void *data)
{
	uint16_t mid = 0;
	int ret;
	struct cmdu_buff *frm;

	trace("%s: %d\n", __func__, __LINE__);
	frm = cmdu_alloc_simple(CMDU_ASSOCIATION_STATUS_NOTIFICATION, &mid);
	if (!frm) {
		dbg("%s: -ENOMEM\n", __func__);
		return NULL;
	}

	/* send notifiy to Multi-AP Controller
	 * and all the other Multi-AP Agent (TODO)
	 */
	memcpy(frm->origin, a->cfg.cntlr_almac, 6);

	ret = agent_gen_assoc_status_notif(a, frm, num_data, data);
	if (ret)
		goto error;

	cmdu_put_eom(frm);

	return frm;
error:
	cmdu_free(frm);

	return NULL;
}

struct cmdu_buff *agent_gen_oper_channel_response(struct agent *a,
		struct wifi_radio_element *radio,
		uint32_t channel, uint32_t bandwidth, bool all)
{
	uint32_t j;
	uint16_t mid = 0;
	struct cmdu_buff *cmdu;
	int ret;

	cmdu = cmdu_alloc_simple(CMDU_OPERATING_CHANNEL_REPORT, &mid);
	if (!cmdu) {
		dbg("%s: -ENOMEM\n", __func__);
		return NULL;
	}

	if (all == 1) {
		/* Operating Channel Report TLV 17.2.17 */
		for (j = 0; j < a->num_radios; j++) {
			struct wifi_radio_element *r = a->radios + j;
			uint32_t chan = r->current_channel;
			uint32_t op_class = r->current_opclass;
			ret = agent_gen_oper_channel_report(a, cmdu, r, chan, 0, op_class);
			if (ret)
				goto error;
		}
	} else {
		ret = agent_gen_oper_channel_report(a, cmdu, radio, channel, bandwidth, 0);
		if (ret)
			goto error;
	}

	/* destination controller */
	memcpy(cmdu->origin, a->cntlr_almac, 6);

	cmdu_put_eom(cmdu);
	return cmdu;

error:
	cmdu_free(cmdu);
	return NULL;
}

struct cmdu_buff *agent_gen_higher_layer_data(struct agent *a, uint8_t *addr,
		uint8_t proto, uint8_t *data, int len)
{
	struct cmdu_buff *frm;
	int ret;

	/* HLD payload with 1 byte proto + tlv header + eom */
	frm = cmdu_alloc_frame(len + 1 + 2*TLV_HLEN);
	if (!frm)
		return NULL;
	cmdu_set_type(frm, CMDU_HIGHER_LAYER_DATA);
	//cmdu_set_mid(frm, 0); /* dynamicly assigned */

	/* Agent send to orgin */
	memcpy(frm->origin, addr, 6);

	ret = agent_gen_tlv_higher_layer_data(a, frm, proto, data, len);
	if (ret)
		goto error;

	cmdu_put_eom(frm);
	return frm;

error:
	cmdu_free(frm);
	return NULL;
}

struct cmdu_buff *agent_gen_sta_caps_response(struct agent *a,
		struct cmdu_buff *rx_cmdu)
{
	uint8_t result_code = 0x00;
	uint8_t reason_code = 0x00;
	uint16_t mid = 0;
	int ret;
	struct cmdu_buff *resp;
	struct sta *s = NULL;
	struct tlv_client_info *tmp = NULL;
	struct tlv *tv[1][16];
	struct tlv_policy cap_policy[] = {
		[0] = {
			.type = MAP_TLV_CLIENT_INFO,
			.present = TLV_PRESENT_ONE
		}
	};

	/* parse received CMDU */
	cmdu_parse_tlvs(rx_cmdu, tv, cap_policy, 1);
	if (!tv[0][0])
		return NULL;

	mid = cmdu_get_mid(rx_cmdu);
	resp = cmdu_alloc_simple(CMDU_CLIENT_CAPABILITY_REPORT, &mid);
	if (!resp) {
		dbg("%s: -ENOMEM\n", __func__);
		return NULL;
	}

	memcpy(resp->origin, rx_cmdu->origin, 6);

	tmp = (struct tlv_client_info *)tv[0][0]->data;

	/* Client Info TLV */
	ret = agent_gen_client_info(a, resp, tmp->macaddr, tmp->bssid);
	if (ret)
		goto error;

	s = find_sta_by_mac(a, tmp->macaddr);
	if (s && (s->assoc_frame)) {
		/* sta is associated */
		trace("sta associated\n");

		/* Client Capability Report TLV */
		ret = agent_gen_client_cap_report(a, resp,
				result_code, s);
		if (ret)
			goto error;

	} else {
		result_code = 0x01;

		if (s) {
			/* assoc frame is not available */
			trace("no assoc frame for sta " MACFMT "\n",
					MAC2STR(tmp->macaddr));
			reason_code = 0x03;
		} else {
			/* sta is not associated with any BSS*/
			trace("sta not associated\n");
			reason_code = 0x02;
		}

		/* Client Capability Report TLV */
		ret = agent_gen_client_cap_report(a, resp,
				result_code, NULL);
		if (ret)
			goto error;

		/* Error Code TLV */
		ret = agent_gen_tlv_error_code(a, resp,
				tmp->macaddr, reason_code);
		if (ret)
			goto error;
	}

	cmdu_put_eom(resp);

	return resp;
error:
	cmdu_free(resp);

	return NULL;
}

struct cmdu_buff *agent_gen_topology_discovery(struct agent *a)
{
	struct cmdu_buff *frm = NULL;
	uint16_t mid = 0;
	uint8_t origin[6] = {0x01, 0x80, 0xc2, 0x00, 0x00, 0x13};
	int ret = 0;

	frm = cmdu_alloc_simple(CMDU_TYPE_TOPOLOGY_DISCOVERY, &mid);
	if (!frm) {
		dbg("%s: -ENOMEM\n", __func__);
		return NULL;
	}

	ret = agent_gen_al_mac(a, frm, a->almac);
	if (ret)
		goto out;

	ret = agent_gen_mac(a, frm, a->almac);
	if (ret)
		goto out;

	memcpy(frm->origin, origin, 6);
	cmdu_put_eom(frm);

	return frm;
out:
	cmdu_free(frm);
	return NULL;
}
